# This example defines an integration between two online games.
# "Hamster Wars" and "Artichoke World," in which a character
# is passed from one game-world to another. 
# An XML message is sent over JMS, illustrating 
# JRuby's role as glue code in distributed Java systems. 
# The message is received as XML, then transformed to a Ruby object, 
# by dynamic code which is fully independent of the static class types.

# Installation note: 
#
# You must add JMS client libraries and configuration to the system classpath, 
# (with JBoss, add all jars in %JBOSS_HOME%\client as well as jndi.properties)
# then run a JMS server with a queue defined. 
# In the JBoss Application server 4.0.4 (with which this was tested),
# add the following to jbossmq-destinations-service.xml
# <mbean code="org.jboss.mq.server.jmx.Queue" 
#        name="jboss.mq.destination:service=Queue,name=HamsterWars"> 
# </mbean> 

# Usage: jruby jms_example [send|receive]'

# The following line is needed for any use of Java from JRuby
require "java"

include_class "javax.jms.Session"
include_class "javax.naming.InitialContext"
include_class "javax.naming.NamingException"
include_class "javax.naming.NameNotFoundException"
# Calling Ruby code in other *.rb files
require "transform"
require "xml_objects"

# For test purposes, this is the name of the origin of the messages
# $ marks a global variable, and the caps marks a constant
$SOURCE_SYSTEM_NAME= "HamsterWars"

def get_test_message_content()
    File.open($SOURCE_SYSTEM_NAME+".xml") { |file| 
        file.readlines.join 
    }
end		
		
def is_queue_defined(jndi_name) 
	begin 
	    InitialContext.new.lookup(jndi_name)
	rescue NameNotFoundException => ne # compare Java's 'catch'
	    print jndi_name, " not found\n"
	    return false
    end  
    return true
end

# Send an XML message over JMS
def send_message
    jms_call{ |connection, queue| # Pipes mark arguments to code block 
        sender_session = connection.createSession(false, Session::AUTO_ACKNOWLEDGE)
        producer = sender_session.createProducer(queue)
        msg = sender_session.createTextMessage(get_test_message_content())
        puts "Sent message"
        producer.send(msg)
    }
end

# Receive an XML message over JMS, then transform it according
# to configuration in transformation.rb
def receive_message 
    jms_call{ |connection, queue|
        receiver_session = connection.createSession(false, Session::AUTO_ACKNOWLEDGE)
        consumer = receiver_session.createConsumer(queue)
        connection.start()
        received_message = consumer.receive(1000)
        raise "no message received after 1 sec" unless received_message
        received_text = received_message.getText()
        
        # Asserting that we received the same XML string that was sent
        raise "Failed: #{get_test_message_content()} != #{received_text} " unless get_test_message_content()  == received_text 

        # Convert XML to a Ruby object, with code that is independent of the object's class
        received_object =  xmlstring_to_object(queue.getName, received_text)
        # Transform document: See xml_objects.rb for details
        target = transform(received_object)
        puts "Successfully received XML for object of type #{received_object.class}\n" + 
        	 "    and transformed to object of type #{target.class}"
        
        # As a follow-on, we could process the resulting object, pass it to another system, or serialize it to XML
        # ....
    }
end

# This executes any arbitrary code block in a context where a JMS
# connection is open, then guarantees the closer of the connection.
def jms_call()
    jndi_name = "/queue/"+$SOURCE_SYSTEM_NAME

    raise "Queue #{$SOURCE_SYSTEM_NAME} must exist" unless is_queue_defined(jndi_name)

    initial_context = InitialContext.new
    connection_factory = initial_context.lookup("/ConnectionFactory")
    queue = initial_context.lookup(jndi_name)

    begin # We make sure to open/close JMS connection, wrapping around an arbitrary code block
        connection= connection_factory.createConnection()
        yield(connection, queue) # Call  code block giving it these arguments
    ensure # comparable to Java 'finally'
        initial_context.close() if initial_context # note 'if' positioned after the then-clause
        connection.close() if connection
    end

end


def run_example(send_or_receive) 
    if send_or_receive.casecmp("send")==0 then # case-insensitive comparison
        send_message
    elsif send_or_receive.casecmp("receive")==0 
        receive_message
    else
        # 'raise' comparable to Java 'throw'
        raise 'Usage: jruby jms_example [send|receive]'
    end
    
end


# ARGV as argument to Java's 'main' method
# if $0 == __FILE__ tests that the file is being executed 
# as a script rather than loaded into another script 
run_example(ARGV[0]) if $0 == __FILE__