# This script transforms a source object to a target object, using
# transformations given in a separate transformation file. The target class
# is generated on the fly based on the information in the transformation file.

require 'transformations' # The configuration file giving the transformations.


def transform(source)
	preprocessed_source = pre_transform(source)
	# The target class is generated on-the-fly
	target_class =get_target_class(preprocessed_source)
	
	# Define a new singleton method -- defined on the source 
	# object alone. This allows user-provided expression to be evaluated 
	# with the source object as the implicit self ('this').
	# The singleton method runs the transformation and returns
	# the target object
	def preprocessed_source.local_transform(target_class)
		target = target_class.new
		# Get all the methods of the target object, 
		# select only those used for assignment (ending in '=', comparable to JavaBean
		# setter methods), and sort them for convenience.
		target.methods.select{ |method_name| 
				method_name.ends_with?("=") and !["==", "==="].include?(method_name) 
			}.sort.each{ |target_method_name| 
			# Call each "setter" method, with the values calculated from the source object.
			begin 
				# Get the "setter" method name without the '='.
				# The [0...-1] syntax takes every character from the beginning of the string through one before the end.
				base_method_name = target_method_name[0...-1]
				expression = get_transformation_expression(target, self, base_method_name)
				trg_value = eval(expression)
				
				if trg_value then 
					target.send(target_method_name, trg_value) 
				end

			rescue Exception # ''rescue' --compare Java's "catch"
				# do nothing, but continue iterating
				print "Error: " + $!
			end
		}
		
		return target
	end
	
	target = preprocessed_source.local_transform(target_class)
	# removing method from 'source' object after using it
	class << source 
    		remove_method(:local_transform) 
    end
    return post_transform(target)
   
end


# Get the transformation (a map assigning the fields in the target class)
# from the transformation-configuration file.
def get_transformation(source)
	source_class_name = source.class.to_s
 	transformation = $TRANSFORMATION[source_class_name]
	raise "Transformation for #{source_class_name} not found in configuration"  unless transformation
	return transformation
end

# Get the class for the target of the transformation, given the source.
# The class is not defined until runtime, when it is dynamically generated
# from the configuration of the transformation.
def get_target_class(source)  
	class_name= get_transformation(source)[:target]
	raise "No transformation target class found for #{source.class}" unless class_name
	
	names = get_transformation(source)[:mappings].keys
	# For a tutorial on this metaprogramming technique,, see  
	# http://www.devsource.com/article2/0,1895,1928561,00.asp, 
	# "An Exercise in Metaprogramming with Ruby", by Hal Fulton
	if Object.const_defined? class_name.to_sym then # Don't define class twice
		klass = Object.const_get(class_name)
	else	
		# Create a new class.
		klass = Object.const_set(class_name, Class.new)
		
		# Add accessor methods
		klass.class_eval do
		      attr_accessor *names
		
		      define_method(:initialize) do |*values|
		        names.each_with_index do |name,i|
		          instance_variable_set("@" + name, values[i])
		        end
		      end
		end
	end

	return klass
end


# Get the transformation expression (which will be calculated, and the value
# assigned using the relevant method in the target object).
def get_transformation_expression(target, source, target_method_name)
	mappings= get_transformation(source)[:mappings]  
	expression = mappings[target_method_name] # can return nil
end


# For convenience, here we add a new method to the built-in class String
class String
  def ends_with?(s)
  	s == self[-s.length, s.length]
  end
end

def test_build_obj_in_memory()
	require "hamster_wars"
	source = HamsterWars::HamsterCharacter.new
	source.customer_first_name = "Freddy"
	source.customer_last_name = "Kane"
	source.character_alias = "Morgoth"
	source.experience = 16000
	source.balance_dollars = 325
	source.hits = 84
 
	target = transform(source)
	assert_expected(target)
end       



def test_obj_from_xml
	require "xml_objects"     
 	source= xmlfile_to_object("HamsterWars.xml")
	target = transform(source)
	assert_expected(target)
end     

# Used by the test methods
def assert_expected(target)
	raise "#{target.customer_full_name} wrong" unless target.customer_full_name == "Kane, Freddy"
	raise "#{target.ingame_name} wrong" unless target.ingame_name == "Morgoth"
	raise "#{target.balance_euros} wrong" unless target.balance_euros == 257.6
	raise "#{target.level} wrong" unless target.level == 4
	raise "#{target.health} wrong" unless target.health == 16
	puts "Transformation test succeeded"
end


# If this file is executed directly, some simple 
# test methods will be run.
if $0 == __FILE__
	test_build_obj_in_memory
	test_obj_from_xml
end

